// Copyright (c) 2009, 2010 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


#ifndef MCT__IMPL__CLOSED_HASH_TABLE_HPP
#define MCT__IMPL__CLOSED_HASH_TABLE_HPP


#include <cmath>
#include <cstddef>
#include <iterator>
#include <memory>
#include <stdexcept>
#include <utility>

#include <mct/impl/optimization.hpp>
#include <mct/impl/utils.hpp>


// Internally data is stored in an array of buckets with length being some power of 2.
// Tables come in two quite different variants: those where buckets keep hash values of
// their keys and those where buckets don't.
//
// Hashes are kept in buckets:
// - each bucket has an additional field '_hash';
// - empty buckets have '_hash' set to 0, debris [*] have it set to 1;
// - hashes are always 'postprocessed' after they are computed by the actual hash
//   function so that hashes in used buckets are never 0 or 1;
// - this mode can be used with any key type.
//
// Hashes are recomputed each time:
// - for empty/debris buckets 0 or 1 is stored where key would be (we have to use some
//   reinterpret_cast black magic for this);
// - at most two buckets are designated as 'special': those are the buckets that have
//   keys that match the marks above, yet contain real values; such buckets can be located
//   anywhere in the table or be not present at all;
// - can only be used for 'small' keys: at most sizeof (long long).
//
// Default is to use the second mode (smaller memory footprint) if key size permits.
//
// If you try to improve this remember that key equality is not the same as bitwise
// equality, especially for structures and other custom types.  Even for plain integral
// type one could specify a different comparator function.  Additionally, you can't call
// hash and comparison functions on non-inserted data: they could assert something or
// dereference an invalid pointer as a result.
//
// [*] Debris are buckets that contained some value but were subsequently erased.  With
//     quadratic probing such buckets cannot be just reset to empty state.

// It was decided that tables never shrink on their own.  See documentation for rationale.


namespace mct
{

  namespace impl
  {

    template <typename Buckets, std::size_t key_type_size>
    struct bucket_without_hash_type_picker
    {
      // We have to store hash except for just a few key type sizes.
      typedef  typename Buckets::with_hash  type;
    };

    template <std::size_t key_type_size>
    struct key_storage_type_picker;

    template <typename Buckets>
    struct bucket_without_hash_type_picker <Buckets, MCT_SIZEOF_CHAR>
    {
      typedef  typename Buckets::without_hash  type;
    };

    template <>
    struct key_storage_type_picker <MCT_SIZEOF_CHAR>
    {
      typedef  unsigned char  type;
    };

# if MCT_SIZEOF_SHORT > MCT_SIZEOF_CHAR

    template <typename Buckets>
    struct bucket_without_hash_type_picker <Buckets, MCT_SIZEOF_SHORT>
    {
      typedef  typename Buckets::without_hash  type;
    };

    template <>
    struct key_storage_type_picker <MCT_SIZEOF_SHORT>
    {
      typedef  unsigned short  type;
    };

#endif

# if MCT_SIZEOF_INT > MCT_SIZEOF_SHORT

    template <typename Buckets>
    struct bucket_without_hash_type_picker <Buckets, MCT_SIZEOF_INT>
    {
      typedef  typename Buckets::without_hash  type;
    };

    template <>
    struct key_storage_type_picker <MCT_SIZEOF_INT>
    {
      typedef  unsigned int  type;
    };

#endif

# if MCT_SIZEOF_LONG > MCT_SIZEOF_INT

    template <typename Buckets>
    struct bucket_without_hash_type_picker <Buckets, MCT_SIZEOF_LONG>
    {
      typedef  typename Buckets::without_hash  type;
    };

    template <>
    struct key_storage_type_picker <MCT_SIZEOF_LONG>
    {
      typedef  unsigned long  type;
    };

#endif

# if MCT_HAVE_LONG_LONG && MCT_SIZEOF_LONG_LONG > MCT_SIZEOF_LONG

    template <typename Buckets>
    struct bucket_without_hash_type_picker <Buckets, MCT_SIZEOF_LONG_LONG>
    {
      typedef  typename Buckets::without_hash  type;
    };

    template <>
    struct key_storage_type_picker <MCT_SIZEOF_LONG_LONG>
    {
      typedef  unsigned long long  type;
    };

#endif


    template <typename Buckets, bool keep_hashes>
    struct bucket_type_picker;

    template <typename Buckets>
    struct bucket_type_picker <Buckets, false>
    {
      typedef  typename bucket_without_hash_type_picker
                 <Buckets, sizeof (typename Buckets::with_hash::key_type)>::type
               type;
    };

    template <typename Buckets>
    struct bucket_type_picker <Buckets, true>
    {
      typedef  typename Buckets::with_hash  type;
    };


    template <typename Traits, bool keep_hashes>
    class bucket_base;


    template <typename Traits>
    class bucket_base <Traits, false>
    {
    public:

      typedef  typename Traits::value_type             value_type;
      typedef  typename Traits::assignable_value_type  assignable_value_type;
      typedef  typename Traits::key_type               key_type;

      static  const bool  KEEPS_HASHES = false;


    private:

      typedef  typename key_storage_type_picker <sizeof (key_type)>::type  key_storage_type;

      char  _value[sizeof (value_type)];


    public:

      static  const key_type&
      extract_key (const value_type& value)
      {  return Traits::extract_key (value);  }

      const key_type&
      key () const
      {  return Traits::extract_key (value ());  }

      const value_type&
      value () const
      {  return *reinterpret_cast <const value_type*> (_value);  }

      value_type&
      value ()
      {  return *reinterpret_cast <value_type*> (_value);  }

      template <typename Allocator>
      void
      construct_value (Allocator& allocator, const value_type& initializer)
      {  allocator.construct (&value (), initializer);  }


      template <typename Hash>
      std::size_t
      key_hash (const Hash& hasher) const
      {
        return hasher (Traits::extract_key (value ()));
      }

      static  std::size_t
      postprocess_hash (std::size_t hash)
      {  return hash;  }


      bool
      used_and_matches (std::size_t hash) const
      {  return !is_empty_or_debris ();  }

      bool
      is_empty () const
      {  return key_storage () == 0;  }

      bool
      is_debris () const
      {  return key_storage () == 1;  }

      bool
      is_empty_or_debris () const
      {  return key_storage () <= 1;  }


      void
      store_hash (std::size_t hash)
      { }

      void
      copy_hash (const bucket_base& that)
      { }

      void
      clear ()
      {
        const_cast <key_storage_type&> (key_storage ()) = 0;
      }

      void
      destroy ()
      {
        const_cast <key_storage_type&> (key_storage ()) = 1;
      }


      void
      revert_to_empty ()
      {  clear ();  }

      void
      revert_to_debris ()
      {  destroy ();  }


    private:

      const key_storage_type&
      key_storage () const
      {
        return reinterpret_cast <const key_storage_type&> (Traits::extract_key (value ()));
      }
    };


    template <typename Traits>
    class bucket_base <Traits, true>
    {
    public:

      typedef  typename Traits::value_type             value_type;
      typedef  typename Traits::assignable_value_type  assignable_value_type;
      typedef  typename Traits::key_type               key_type;

      static  const bool  KEEPS_HASHES = true;


    private:

      std::size_t  _hash;
      value_type   _value;


    public:

      static  const key_type&
      extract_key (const value_type& value)
      {  return Traits::extract_key (value);  }

      const key_type&
      key () const
      {  return Traits::extract_key (_value);  }

      const value_type&
      value () const
      {  return _value;  }

      value_type&
      value ()
      {  return _value;  }

      template <typename Allocator>
      void
      construct_value (Allocator& allocator, const value_type& initializer)
      {  allocator.construct (&_value, initializer);  }


      template <typename Hash>
      std::size_t
      key_hash (const Hash& hasher) const
      {  return _hash;  }

      static  std::size_t
      postprocess_hash (std::size_t hash)
      {
        return MCT_OPTIMIZATION_LIKELY (hash >= 2) ? hash : hash + 2;
      }


      // 'hash' is guaranteed to be different from empty/debris mark, so just comparing
      // with that is both 'used' and 'matches' test at the same time.
      bool
      used_and_matches (std::size_t hash) const
      {  return this->_hash == hash;  }

      bool
      is_empty () const
      {  return _hash == 0;  }

      bool
      is_debris () const
      {  return _hash == 1;  }

      bool
      is_empty_or_debris () const
      {  return _hash <= 1;  }


      void
      store_hash (std::size_t hash)
      {  this->_hash = hash;  }

      void
      copy_hash (const bucket_base& that)
      {  this->_hash = that._hash;  }

      void
      clear ()
      {  _hash = 0;  }

      void
      destroy ()
      {  _hash = 1;  }


      // Don't do anything because 'hash' field is not changed anyway.
      void
      revert_to_empty ()
      { }

      void
      revert_to_debris ()
      { }
    };


    // Note that 'buckets' and 'num_buckets' are not initialized.
    template <typename Bucket, bool with_special_buckets = !Bucket::KEEPS_HASHES>
    struct closed_hash_table_data_base;


    template <typename Bucket>
    struct closed_hash_table_data_base <Bucket, false>
    {
      typedef  Bucket  bucket_type;

      bucket_type*  buckets;
      std::size_t   num_buckets;
      std::size_t   num_used;
      std::size_t   num_occupied;


      closed_hash_table_data_base ()
        : num_used     (0),
          num_occupied (0)
      { }


      bool
      is_special (bucket_type* bucket) const
      {  return false;  }

      void
      note_new (bucket_type* bucket)
      { }

      void
      note_being_destroyed (bucket_type* bucket)
      { }


      void
      clear ()
      {
        num_used     = 0;
        num_occupied = 0;
      }


      void
      swap (closed_hash_table_data_base& that)
      {
        std::swap (buckets,      that.buckets);
        std::swap (num_buckets,  that.num_buckets);
        std::swap (num_used,     that.num_used);
        std::swap (num_occupied, that.num_occupied);
      }
    };


    template <typename Bucket>
    struct closed_hash_table_data_base <Bucket, true>
      : public closed_hash_table_data_base <Bucket, false>
    {
      typedef  closed_hash_table_data_base <Bucket, false>  base_type;

      typedef  Bucket                                       bucket_type;


      bucket_type*  special_buckets[2];


      closed_hash_table_data_base ()
      {
        special_buckets[0] = 0;
        special_buckets[1] = 0;
      }


      bool
      is_special (bucket_type* bucket) const
      {
        return MCT_OPTIMIZATION_UNLIKELY (bucket    == special_buckets[0]
                                          || bucket == special_buckets[1]);
      }


      void
      note_new (bucket_type* bucket)
      {
        if (MCT_OPTIMIZATION_UNLIKELY (bucket->is_empty_or_debris ()))
          special_buckets[bucket->is_empty () ? 0 : 1] = bucket;
      }

      void
      note_being_destroyed (bucket_type* bucket)
      {
        if (MCT_OPTIMIZATION_UNLIKELY (bucket->is_empty_or_debris ()))
          special_buckets[bucket->is_empty () ? 0 : 1] = 0;
      }


      void
      clear ()
      {
        base_type::clear ();
        special_buckets[0] = 0;
        special_buckets[1] = 0;
      }


      void
      swap (closed_hash_table_data_base& that)
      {
        base_type::swap (that);

        // At least VC 2005 cannot swap arrays, bug #549067.
        std::swap (special_buckets[0], that.special_buckets[0]);
        std::swap (special_buckets[1], that.special_buckets[1]);
      }
    };


    template <typename Bucket>
    struct closed_hash_table_data : public closed_hash_table_data_base <Bucket>
    {
      typedef  closed_hash_table_data_base <Bucket>  base_type;

      typedef  Bucket                                bucket_type;


      std::size_t
      start_probing (std::size_t hash) const
      {
        // We require that number of buckets is a power of 2.  To compensate, we involve a
        // prime number here; this usually improves speed with particularly bad hash
        // function choices.
#     if MCT_SIZEOF_STD_SIZE_T == 4
        return (hash * 0x3874e607) & (base_type::num_buckets - 1);
#     else
        return (hash * 0x1a4548ec34f6e06f) & (base_type::num_buckets - 1);
#     endif
      }

      std::size_t
      continue_probing (std::size_t looking_at, std::size_t iteration) const
      {
        // Quadratic probing.
        return (looking_at + iteration) & (base_type::num_buckets - 1);
      }


      bucket_type*
      end () const
      {
        return base_type::buckets + base_type::num_buckets;
      }


      bucket_type*  first_used_bucket () const;
      bucket_type*  next_used_bucket     (bucket_type* bucket)  const;
      bucket_type*  previous_used_bucket (bucket_type* bucket)  const;
    };


    template <typename Bucket>
    Bucket*
    closed_hash_table_data <Bucket>::
    first_used_bucket () const
    {
      bucket_type* const  _end = end ();

      if (base_type::num_used != 0)
        {
          for (bucket_type* bucket = base_type::buckets; bucket != _end; ++bucket)
            {
              if (!bucket->is_empty_or_debris () || base_type::is_special (bucket))
                return bucket;
            }
        }

      return _end;
    }

    template <typename Bucket>
    Bucket*
    closed_hash_table_data <Bucket>::
    next_used_bucket (bucket_type* bucket) const
    {
      for (bucket_type* const _end = end (); ++bucket != _end;)
        {
          if (!bucket->is_empty_or_debris () || base_type::is_special (bucket))
            break;
        }

      return bucket;
    }

    template <typename Bucket>
    Bucket*
    closed_hash_table_data <Bucket>::
    previous_used_bucket (bucket_type* bucket) const
    {
      // There is no out-of-bounds state in the beginning.
      while (true)
        {
          --bucket;
          if (!bucket->is_empty_or_debris () || base_type::is_special (bucket))
            return bucket;
        }
    }


    template <typename Bucket>
    class hash_table_const_iterator
    {
      typedef  Bucket  bucket_type;

    protected:

      bucket_type*                                 _bucket;
      const closed_hash_table_data <bucket_type>*  _data;

    public:

      typedef  std::bidirectional_iterator_tag         iterator_category;
      typedef  const typename bucket_type::value_type  value_type;
      typedef  std::ptrdiff_t                          difference_type;
      typedef  std::size_t                             size_type;
      typedef  value_type*                             pointer;
      typedef  value_type&                             reference;


      hash_table_const_iterator ()
      { }

      hash_table_const_iterator (bucket_type* bucket,
                                 const closed_hash_table_data <bucket_type>* data)
        : _bucket (bucket),
          _data   (data)
      { }


      value_type&
      operator* () const
      {  return _bucket->value ();  }

      value_type*
      operator-> () const
      {  return &_bucket->value ();  }


      hash_table_const_iterator&
      operator++ ()
      {
        _bucket = _data->next_used_bucket (_bucket);
        return *this;
      }

      hash_table_const_iterator
      operator++ (int)
      {
        const hash_table_const_iterator  copy (*this);

        _bucket = _data->next_used_bucket (_bucket);
        return copy;
      }

      hash_table_const_iterator&
      operator-- ()
      {
        _bucket = _data->previous_used_bucket (_bucket);
        return *this;
      }

      hash_table_const_iterator
      operator-- (int)
      {
        const hash_table_const_iterator  copy (*this);

        _bucket = _data->previous_used_bucket (_bucket);
        return copy;
      }


      // Supposedly not for outsiders to use.

      bucket_type*
      bucket () const
      {  return _bucket;  }

      const closed_hash_table_data <bucket_type>*
      data () const
      {  return _data;  }
    };


    template <typename Bucket>
    class hash_table_iterator : public hash_table_const_iterator <Bucket>
    {
      typedef  Bucket                                   bucket_type;
      typedef  hash_table_const_iterator <bucket_type>  base_type;

    public:

      typedef  std::bidirectional_iterator_tag              iterator_category;
      typedef  typename bucket_type::assignable_value_type  value_type;
      typedef  ptrdiff_t                                    difference_type;
      typedef  value_type*                                  pointer;
      typedef  value_type&                                  reference;


      hash_table_iterator ()
      { }

      hash_table_iterator (bucket_type* bucket, const closed_hash_table_data <bucket_type>* data)
        : base_type (bucket, data)
      { }


      value_type&
      operator* () const
      {  return base_type::_bucket->value ();  }

      value_type*
      operator-> () const
      {  return &base_type::_bucket->value ();  }


      hash_table_iterator&
      operator++ ()
      {
        return static_cast <hash_table_iterator&> (base_type::operator++ ());
      }

      hash_table_iterator
      operator++ (int)
      {
        return static_cast <hash_table_iterator> (base_type::operator++ ());
      }

      hash_table_iterator&
      operator-- ()
      {
        return static_cast <hash_table_iterator&> (base_type::operator-- ());
      }

      hash_table_iterator
      operator-- (int)
      {
        return static_cast <hash_table_iterator> (base_type::operator-- ());
      }
    };


    template <typename Bucket>
    inline  bool
    operator== (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () == iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator!= (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () != iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator< (const hash_table_const_iterator <Bucket>& iterator1,
               const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () < iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator<= (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () <= iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator> (const hash_table_const_iterator <Bucket>& iterator1,
               const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () > iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator>= (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () >= iterator2.bucket ();  }


    const std::size_t  DEFAULT_NUM_BUCKETS     = 16;
    const std::size_t  MIN_NUM_BUCKETS         = 4;
    const float        DEFAULT_MAX_LOAD_FACTOR = 0.6f;
    const float        MIN_MAX_LOAD_FACTOR     = 0.01f;
    const float        MAX_MAX_LOAD_FACTOR     = 0.99f;


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    class closed_hash_table
    {
      template <typename _Buckets, typename _Hash, typename _Equal, typename _Allocator,
                bool _keep_hashes>
      friend  bool
      operator== (const closed_hash_table <_Buckets, _Hash, _Equal, _Allocator,
                                           _keep_hashes>& table1,
                  const closed_hash_table <_Buckets, _Hash, _Equal, _Allocator,
                                           _keep_hashes>& table2);


    protected:

      typedef  typename bucket_type_picker <Buckets, keep_hashes>::type  bucket_type;
      typedef  closed_hash_table_data <bucket_type>                      data_type;

    public:

      typedef  typename bucket_type::key_type                            key_type;
      typedef  typename bucket_type::value_type                          value_type;
      typedef  Hash                                                      hasher;
      typedef  Equal                                                     key_equal;
      typedef  typename Allocator::template rebind <value_type>::other   allocator_type;

      typedef  hash_table_iterator <bucket_type>                         iterator;
      typedef  hash_table_const_iterator <bucket_type>                   const_iterator;

      typedef  typename allocator_type::pointer                          pointer;
      typedef  typename allocator_type::const_pointer                    const_pointer;
      typedef  typename allocator_type::reference                        reference;
      typedef  typename allocator_type::const_reference                  const_reference;
      typedef  std::ptrdiff_t                                            difference_type;
      typedef  std::size_t                                               size_type;


# if MCT_ENABLE_DEBUGGING

      struct statistics
      {
        double     debris_ratio;
        double     avg_present_lookup;
        size_type  max_present_lookup;
        double     avg_absent_lookup;
        size_type  max_absent_lookup;
      };

# endif


    protected:

      typedef  typename allocator_type::template rebind <bucket_type>::other
               bucket_allocator_type;


      data_type       _data;
      size_type       _max_occupied;
      hasher          _hash;
      key_equal       _equal;
      allocator_type  _allocator;
      float           _max_load_factor;


      // This structure is used to simplify exception safety code.
      struct data_grip : public data_type
      {
        closed_hash_table&  table;

        data_grip (closed_hash_table& table, size_type num_buckets)
          : table (table)
        {
          table.do_allocate_buckets (*this, num_buckets);
        }

        ~data_grip ()
        {
          table.do_deallocate_buckets (*this);
        }
      };


    public:

      closed_hash_table (size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : _hash      (hash),
          _equal     (equal),
          _allocator (allocator)
      {
        allocate_buckets (num_buckets);
        do_set_max_load_factor (DEFAULT_MAX_LOAD_FACTOR);
      }

      closed_hash_table (const closed_hash_table& that);
      closed_hash_table (const closed_hash_table& that, const allocator_type& allocator);

      template <typename InputIterator>
      closed_hash_table (InputIterator first, InputIterator last,
                         size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator);

    private:

      void  initialize_as_copy (const closed_hash_table& that);

    public:

      ~closed_hash_table ();


      closed_hash_table&  operator= (const closed_hash_table& that);


      iterator
      begin ()
      {
        return make_iterator (_data.first_used_bucket ());
      }

      const_iterator
      begin () const
      {
        return make_const_iterator (_data.first_used_bucket ());
      }

      const_iterator
      cbegin () const
      {
        return make_const_iterator (_data.first_used_bucket ());
      }

      iterator
      end ()
      {
        return make_iterator (_data.end ());
      }

      const_iterator
      end () const
      {
        return make_const_iterator (_data.end ());
      }

      const_iterator
      cend () const
      {
        return make_const_iterator (_data.end ());
      }


      bool
      empty () const
      {  return _data.num_used == 0;  }

      size_type
      size () const
      {  return _data.num_used;  }

      size_type
      max_size () const
      {
        return static_cast <size_type> (std::floor (max_bucket_count () * MAX_MAX_LOAD_FACTOR));
      }


      float
      load_factor () const
      {
        return static_cast <float> (_data.num_used) / _data.num_buckets;
      }

      float
      max_load_factor () const
      {  return _max_load_factor;  }

      void
      max_load_factor (float max_load_factor)
      {
        do_set_max_load_factor (max_load_factor);
        if (_data.num_occupied > _max_occupied)
          clear_debris_or_grow ();
      }


      size_type
      bucket_count () const
      {  return _data.num_buckets;  }

      size_type
      max_bucket_count () const
      {
        return round_down_to_power_of_2 (bucket_allocator_type (_allocator).max_size ());
      }


      hasher
      hash_function () const
      {  return _hash;  }

      key_equal
      key_eq () const
      {  return _equal;  }

      allocator_type
      get_allocator () const
      {  return _allocator;  }


      iterator
      find (const key_type& key)
      {
        return make_iterator (lookup (key));
      }

      const_iterator
      find (const key_type& key) const
      {
        return make_const_iterator (lookup (key));
      }

      size_type
      count (const key_type& key) const
      {
        return lookup (key) != _data.end () ? 1 : 0;
      }

      std::pair <iterator, iterator>
      equal_range (const key_type& key)
      {
        const iterator  first (make_iterator (lookup (key)));
        iterator        last  (first);

        if (first.bucket () != _data.end ())
          ++last;

        return std::make_pair (first, last);
      }

      std::pair <const_iterator, const_iterator>
      equal_range (const key_type& key) const
      {
        const const_iterator  first (make_const_iterator (lookup (key)));
        const_iterator        last  (first);

        if (first.bucket () != _data.end ())
          ++last;

        return std::make_pair (first, last);
      }


      std::pair <iterator, bool>
      insert (const value_type& value)
      {
        const std::pair <bucket_type*, bool>  result (do_insert (value));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      // Note that we currently just ignore 'hint'.
      iterator
      insert (const_iterator hint, const value_type& value)
      {
        return make_iterator (do_insert (value).first);
      }

      template <typename InputIterator>
      void
      insert (InputIterator first, InputIterator last)
      {
        const size_type  num_inserted = forward_distance (first, last);
        if (num_inserted && _data.num_used + num_inserted > _max_occupied)
          rehash (static_cast <size_type> ((_data.num_used + num_inserted) / max_load_factor ()));

        for (; first != last; ++first)
          do_insert (*first);
      }


      size_type
      erase (const key_type& key)
      {
        return do_erase (key) ? 1 : 0;
      }

      iterator
      erase (const_iterator position)
      {
        do_erase (position.bucket ());
        return make_iterator (_data.next_used_bucket (position.bucket ()));
      }

      iterator
      erase (const_iterator first, const_iterator last)
      {
        bucket_type*  bucket = first.bucket ();
        for (bucket_type* const limit = last.bucket (); bucket != limit;
             bucket = _data.next_used_bucket (bucket))
          do_erase (bucket);

        return make_iterator (bucket);
      }


      void  clear ();
      void  swap (closed_hash_table& that);
      void  rehash (size_type num_buckets);


#   if MCT_ENABLE_DEBUGGING

      bool
      valid_iterator (const_iterator position) const
      {
        // In some unlikely cases certain iterators (e.g. from another table after several
        // rehashes) could point to a valid bucket, but still be invalid.
        if (position.data () != &_data)
          return false;

        bucket_type* const  bucket = position.bucket ();
        return (_data.buckets <= bucket && bucket < _data.end ()
                // Check that it doesn't point into a middle of an actual bucket.
                && ((reinterpret_cast <char*> (bucket) - reinterpret_cast <char*> (_data.buckets))
                    % sizeof (bucket_type) == 0)
                && (!bucket->is_empty_or_debris () || _data.is_special (bucket)));
      }

      size_type
      used_memory () const
      {
        return sizeof *this + _data.num_buckets * sizeof (bucket_type);
      }

      void        validate_integrity () const;
      statistics  collect_statistics () const;

#   endif


    protected:

      iterator
      make_iterator (bucket_type* bucket) const
      {  return iterator (bucket, &_data);  }

      const_iterator
      make_const_iterator (bucket_type* bucket) const
      {  return const_iterator (bucket, &_data);  }


      bucket_type*  lookup (const key_type& key)  const;

      // Templated for convenience of map's operator[].
      template <typename type>
      std::pair <bucket_type*, bool>  lookup_or_insert (const type& data);


      void  do_set_max_load_factor (float max_load_factor);


      std::pair <bucket_type*, bool>
      do_insert (const value_type& value)
      {
        return lookup_or_insert <value_type> (value);
      }


      bool
      do_erase (const key_type& key)
      {
        bucket_type* const  bucket = lookup (key);

        if (bucket != _data.end ())
          {
            do_erase (bucket);
            return true;
          }
        else
          return false;
      }

      void
      do_erase (bucket_type* bucket)
      {
        _data.note_being_destroyed (bucket);
        --_data.num_used;
        _allocator.destroy (&bucket->value ());
        bucket->destroy ();
      }


      void  do_rehash (size_type num_buckets);
      void  clear_debris_or_grow ();


      size_type
      compute_max_occupied () const
      {
        if (_data.num_buckets < max_bucket_count ())
          {
            return std::min (_data.num_buckets - 1,
                             static_cast <size_type> (_data.num_buckets * _max_load_factor));
          }
        else
          return max_size ();
      }


    private:

      void
      allocate_buckets (size_type num_buckets)
      {
        if (num_buckets > MIN_NUM_BUCKETS)
          num_buckets = round_up_to_power_of_2 (num_buckets);
        else
          num_buckets = (num_buckets ? MIN_NUM_BUCKETS : DEFAULT_NUM_BUCKETS);

        do_allocate_buckets (_data, num_buckets);
      }

      void
      do_allocate_buckets (data_type& data, size_type num_buckets)
      {
        if (num_buckets > max_bucket_count ())
          throw std::length_error ("closed_hash_table::do_allocate_buckets");

        data.num_buckets = num_buckets;
        data.buckets     = bucket_allocator_type (_allocator).allocate (num_buckets);

        for (bucket_type* bucket = data.buckets, * const limit = data.end ();
             bucket != limit; ++bucket)
          bucket->clear ();
      }

      void
      do_deallocate_buckets (const data_type& data)
      {
        // Note: we mustn't depend on exact value of 'num_occupied', because that might be
        // incorrect if we are cleaning up after an exception.  However, it will not be
        // zero if the table is or can be non-empty.
        if (!has_trivial_destructor <value_type>::value && data.num_occupied)
          {
            for (bucket_type* bucket = data.buckets, * const limit = data.end ();
                 bucket != limit; ++bucket)
              {
                if (!bucket->is_empty_or_debris () || data.is_special (bucket))
                  _allocator.destroy (&bucket->value ());
              }
          }

        bucket_allocator_type (_allocator).deallocate (data.buckets, data.num_buckets);
      }


      // This is really internal.  Destination bucket array must be completely clean and
      // have enough space to fit the used elements in 'from'.  'from' must have been
      // built using the same equality and hash functions as this table.
      void  copy_buckets (const data_type& from, data_type& to);
    };


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    closed_hash_table (const closed_hash_table& that)
      : _hash            (that.hash_function ()),
        _equal           (that.key_eq ()),
        _allocator       (that.get_allocator ()),
        _max_load_factor (that.max_load_factor ())
    {
      initialize_as_copy (that);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    closed_hash_table (const closed_hash_table& that, const allocator_type& allocator)
      : _hash            (that.hash_function ()),
        _equal           (that.key_eq ()),
        _allocator       (allocator),
        _max_load_factor (that.max_load_factor ())
    {
      initialize_as_copy (that);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    template <typename InputIterator>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    closed_hash_table (InputIterator first, InputIterator last,
                       size_type num_buckets, const hasher& hash, const key_equal& equal,
                       const allocator_type& allocator)
      : _hash      (hash),
        _equal     (equal),
        _allocator (allocator)
    {
      allocate_buckets (num_buckets ? num_buckets : forward_distance (first, last));
      do_set_max_load_factor (DEFAULT_MAX_LOAD_FACTOR);

      try
        {
          for (; first != last; ++first)
            do_insert (*first);
        }
      catch (...)
        {
          do_deallocate_buckets (_data);
          throw;
        }
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    inline  void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    initialize_as_copy (const closed_hash_table& that)
    {
      allocate_buckets (static_cast <size_type> (std::ceil (that._data.num_used
                                                            / max_load_factor ())));
      _max_occupied = compute_max_occupied ();

      try
        {
          copy_buckets (that._data, _data);
        }
      catch (...)
        {
          do_deallocate_buckets (_data);
          throw;
        }
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    ~closed_hash_table ()
    {
      do_deallocate_buckets (_data);
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>&
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    operator= (const closed_hash_table& that)
    {
      // Just an optimization; will work without this check too.
      if (&that != this)
        {
          closed_hash_table  copy (that, _allocator);
          swap (copy);
        }

      return *this;
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    do_set_max_load_factor (float max_load_factor)
    {
      _max_load_factor = std::max (MIN_MAX_LOAD_FACTOR,
                                   std::min (max_load_factor, MAX_MAX_LOAD_FACTOR));
      _max_occupied    = compute_max_occupied ();
    }


    // According to benchmarks, explicitly unrolling first pass gives a small speedup for
    // lookup() and lookup_or_insert().

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::bucket_type*
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    lookup (const key_type& key) const
    {
      const size_type  hash    = bucket_type::postprocess_hash (_hash (key));
      size_type        look_at = _data.start_probing (hash);

      bucket_type* const  bucket = (_data.buckets + look_at);
      if (bucket->used_and_matches (hash) || _data.is_special (bucket))
        {
          if (_equal (bucket->key (), key))
            return bucket;
        }
      else if (bucket->is_empty ())
        return _data.end ();

      for (size_type iteration = 1; ; ++iteration)
        {
          look_at = _data.continue_probing (look_at, iteration);

          bucket_type* const  bucket = (_data.buckets + look_at);
          if (bucket->used_and_matches (hash) || _data.is_special (bucket))
            {
              if (_equal (bucket->key (), key))
                return bucket;
            }
          else if (bucket->is_empty ())
            return _data.end ();
        }
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    template <typename type>
    std::pair <typename closed_hash_table <Buckets, Hash, Equal, Allocator,
                                           keep_hashes>::bucket_type*,
               bool>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    lookup_or_insert (const type& data)
    {
      const key_type&  key     = bucket_type::extract_key (data);
      const size_type  hash    = bucket_type::postprocess_hash (_hash (key));
      size_type        look_at = _data.start_probing (hash);

      bucket_type* const  bucket = (_data.buckets + look_at);
      bucket_type*        debris;

      if (bucket->used_and_matches (hash) || _data.is_special (bucket))
        {
          if (_equal (bucket->key (), key))
            return std::make_pair (bucket, false);

          debris = 0;
        }
      else if (bucket->is_empty ())
        {
          if (_data.num_occupied >= _max_occupied)
            {
              clear_debris_or_grow ();
              return lookup_or_insert <type> (data);
            }

          try
            {
              bucket->construct_value (_allocator, data);
            }
          catch (...)
            {
              bucket->revert_to_empty ();
              throw;
            }

          bucket->store_hash (hash);
          _data.note_new (bucket);

          ++_data.num_used;
          ++_data.num_occupied;

          return std::make_pair (bucket, true);
        }
      else
        debris = (!bucket_type::KEEPS_HASHES || bucket->is_debris () ? bucket : 0);

      for (size_type iteration = 1; ; ++iteration)
        {
          look_at = _data.continue_probing (look_at, iteration);

          bucket_type*  bucket = (_data.buckets + look_at);

          if (bucket->used_and_matches (hash) || _data.is_special (bucket))
            {
              if (_equal (bucket->key (), key))
                return std::make_pair (bucket, false);
            }
          else if (bucket->is_empty ())
            {
              if (!debris)
                {
                  if (_data.num_occupied >= _max_occupied)
                    {
                      clear_debris_or_grow ();
                      return lookup_or_insert <type> (data);
                    }
                }
              else
                bucket = debris;

              try
                {
                  bucket->construct_value (_allocator, data);
                }
              catch (...)
                {
                  if (bucket != debris)
                    bucket->revert_to_empty ();
                  else
                    bucket->revert_to_debris ();

                  throw;
                }

              bucket->store_hash (hash);
              _data.note_new (bucket);

              ++_data.num_used;
              if (bucket != debris)
                ++_data.num_occupied;

              return std::make_pair (bucket, true);
            }
          else if (!debris && (!bucket_type::KEEPS_HASHES || bucket->is_debris ()))
            debris = bucket;
        }
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    clear ()
    {
      if (!_data.num_occupied)
        return;

      for (bucket_type* bucket = _data.buckets, * const limit = _data.end ();
           bucket != limit; ++bucket)
        {
          if (!has_trivial_destructor <value_type>::value)
            {
              if (!bucket->is_empty_or_debris () || _data.is_special (bucket))
                _allocator.destroy (&bucket->value ());
            }

          bucket->clear ();
        }

      _data.clear ();
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    swap (closed_hash_table& that)
    {
      if (this->_allocator == that._allocator)
        _data.swap (that._data);
      else
        {
          // 'new_this_data' must be allocated with 'this' allocator, but data must come
          // from 'that' and also copied by 'that' (because hash function and equality
          // predicate are not swapped yet).
          data_grip  new_this_data (*this, that ._data.num_buckets);
          data_grip  new_that_data (that,  this->_data.num_buckets);

          that .copy_buckets (that. _data, new_this_data);
          this->copy_buckets (this->_data, new_that_data);

          this->_data.swap (new_this_data);
          that ._data.swap (new_that_data);
        }

      std::swap (_max_occupied,    that._max_occupied);
      std::swap (_hash,            that._hash);
      std::swap (_equal,           that._equal);
      std::swap (_max_load_factor, that._max_load_factor);
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    rehash (size_type num_buckets)
    {
      num_buckets = std::max (num_buckets, static_cast <size_type> (_data.num_used
                                                                    / max_load_factor ()));
      num_buckets = round_up_to_power_of_2 (std::max (num_buckets, _data.num_used + 1));

      // If number of buckets isn't different still rehash to get rid of debris.
      if (num_buckets != _data.num_buckets || _data.num_occupied > _data.num_used)
        do_rehash (num_buckets);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    do_rehash (size_type num_buckets)
    {
      data_grip  new_data (*this, num_buckets);

      copy_buckets (_data, new_data);
      _data.swap (new_data);

      _max_occupied = compute_max_occupied ();
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    clear_debris_or_grow ()
    {
      int  num_buckets = _data.num_buckets;

      // Bit shifting trick to avoid overflow if table is *huge*.
      if (_data.num_used >= (_data.num_occupied - (_data.num_occupied >> 4)))
        {
          // If almost all table is filled with real data rather than debris, we increase
          // table size instead.
          num_buckets <<= 1;
        }

      do_rehash (num_buckets);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    copy_buckets (const data_type& from, data_type& to)
    {
      if (!from.num_used)
        return;

      // If copying is interrupted with exception we needn't have correct values here, but
      // we need to at least have non-zero values to bust possible is-empty optimizations.
      to.num_used     = from.num_used;
      to.num_occupied = from.num_used;

      for (bucket_type* bucket = from.buckets, * const limit = (from.buckets + from.num_buckets);
           bucket != limit; ++bucket)
        {
          if (bucket->is_empty_or_debris () && !from.is_special (bucket))
            continue;

          const size_type  hash    = bucket->key_hash (_hash);
          size_type        look_at = to.start_probing (hash);

          bucket_type*  new_bucket = (to.buckets + look_at);

          if (!new_bucket->is_empty () || to.is_special (new_bucket))
            {
              for (size_type iteration = 1; ; ++iteration)
                {
                  look_at = to.continue_probing (look_at, iteration);

                  new_bucket = (to.buckets + look_at);
                  if (new_bucket->is_empty () && !to.is_special (new_bucket))
                    break;
                }
            }

          try
            {
              new_bucket->construct_value (_allocator, bucket->value ());
            }
          catch (...)
            {
              new_bucket->revert_to_empty ();
              throw;
            }

          new_bucket->copy_hash (*bucket);
          to.note_new (new_bucket);
        }
    }


# if MCT_ENABLE_DEBUGGING

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    validate_integrity () const
    {
      if (_max_occupied != compute_max_occupied ())
        throw std::logic_error ("occupation threshold is incorrect");

      if (_data.num_occupied > _max_occupied)
        throw std::logic_error ("too many occupied buckets");

      size_type  num_used     = 0;
      size_type  num_occupied = 0;

      for (bucket_type* bucket = _data.buckets, * const _end = _data.end ();
           bucket != _end; ++bucket)
        {
          if (!_data.is_special (bucket))
            {
              if (bucket->is_empty ())
                continue;

              ++num_occupied;
              if (bucket->is_debris ())
                continue;
            }
          else
            ++num_occupied;

          ++num_used;
          if (lookup (bucket->key ()) != bucket)
            throw std::logic_error ("bucket lookup fails");
        }

      if (_data.num_used != num_used)
        throw std::logic_error ("wrong number of used buckets");

      if (_data.num_occupied != num_occupied)
        throw std::logic_error ("wrong number of occupied buckets");
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::statistics
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    collect_statistics () const
    {
      statistics  stats;

      stats.debris_ratio       = 0;
      stats.avg_present_lookup = 0;
      stats.max_present_lookup = 0;
      stats.avg_absent_lookup  = 0;
      stats.max_absent_lookup  = 0;

      if (_data.num_occupied != 0)
        {
          stats.debris_ratio = (static_cast <double> (_data.num_occupied - _data.num_used)
                                / _data.num_occupied);
        }

      for (bucket_type* bucket = _data.buckets, * const _end = _data.end ();
           bucket != _end; ++bucket)
        {
          if (!bucket->is_empty_or_debris () || _data.is_special (bucket))
            {
              const size_type  hash    = _hash (bucket->key ());
              size_type        look_at = _data.start_probing (hash);

              for (size_type iteration = 1; ; ++iteration)
                {
                  if (_data.buckets + look_at == bucket)
                    {
                      stats.avg_present_lookup += iteration;
                      stats.max_present_lookup  = std::max (stats.max_present_lookup, iteration);
                      break;
                    }

                  look_at = _data.continue_probing (look_at, iteration);
                }
            }

          size_type  look_at = (bucket - _data.buckets);

          for (size_type iteration = 1; ; ++iteration)
            {
              if ((_data.buckets + look_at)->is_empty ())
                {
                  stats.avg_absent_lookup += iteration;
                  stats.max_absent_lookup  = std::max (stats.max_absent_lookup, iteration);
                  break;
                }

              look_at = _data.continue_probing (look_at, iteration);
            }
        }

      if (_data.num_used != 0)
        stats.avg_present_lookup /= _data.num_used;

      stats.avg_absent_lookup /= _data.num_buckets;

      return stats;
    }

# endif


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    bool
    operator== (const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table1,
                const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table2)
    {
      typedef  closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>  table_type;
      typedef  typename table_type::bucket_type  bucket_type;
      typedef  typename table_type::size_type    size_type;

      if (&table1 == &table2)
        return true;

      const size_type  size = table1.size ();
      if (table2.size () != size)
        return false;

      if (size > 0)
        {
          const bucket_type* const             end = table2._data.end ();
          typename table_type::const_iterator  iterator (table1.begin ());

          for (size_type k = 0; k < size; ++k, ++iterator)
            {
              const bucket_type*  match = table2.lookup (iterator.bucket ()->key ());
              if (match == end || !match->mapped_equal (*iterator.bucket ()))
                return false;
            }
        }

      return true;
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    inline  bool
    operator!= (const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table1,
                const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table2)
    {
      return !operator== (table1, table2);
    }

  }

}


#endif  // Multi-inclusion guard.


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
